@@ -120,7 +120,7 @@ char copyright[] =
 #define TIOCPKT_FLUSHWRITE      0x02
 #endif
 
-#ifdef __386BSD__
+#ifdef HAVE_SYS_IOCTL_COMPAT_H
 #include <sys/ioctl_compat.h>
 #endif
 
@@ -585,7 +585,7 @@ main(argc, argv)
       authopts |= OPTS_FORWARDABLE_CREDS;
 
     status = kcmd(&sock, &host, debug_port,
-		  null_local_username ? NULL : pwd->pw_name,
+		  null_local_username ? "" : pwd->pw_name,
 		  name ? name : pwd->pw_name, term,
 		  0, "host", krb_realm,
 		  &cred,
@@ -614,10 +614,10 @@ main(argc, argv)
     }
 #else
     rem = rcmd(&host, debug_port,
-	       null_local_username ? NULL : pwd->pw_name,
+	       null_local_username ? "" : pwd->pw_name,
 	       name ? name : pwd->pw_name, term, 0);
 #endif /* KERBEROS */
-    
+
     if (rem < 0)
       exit(1);
     
